require fflib.fs
require debug.fs

library libobjc libobjc.dylib
library foundation /System/Library/Frameworks/Foundation.framework/Foundation
library appkit /System/Library/Frameworks/AppKit.framework/AppKit

libobjc objc_getClass ptr (ptr) objc_getClass
libobjc objc_addClass ptr (void) objc_addClass
libobjc objc_lookUpClass ptr (ptr) objc_lookUpClass
libobjc objc_addMethods ptr ptr (void) class_addMethods

libobjc objc_msgSend0 ptr ptr (ptr) objc_msgSend
libobjc objc_msgSend1 ptr ptr ptr (ptr) objc_msgSend
libobjc objc_msgSendv ptr ptr int ptr (ptr) objc_msgSendv
libobjc objc_registerName ptr (ptr) sel_registerName

foundation NSLog0 ptr (void) NSLog
foundation NSLog1 ptr ptr (void) NSLog

appkit NSApplicationMain int ptr (int) NSApplicationMain

: ftoc ( c-addr u -- 'c-addr ) here dup >r over allot swap cmove 0 c, align r> ;

: objc_class ( addr u -- ptr )
ftoc objc_getClass ;

: objc_sel ( addr u -- ptr )
ftoc objc_registerName ;

s" NSObject" objc_class Constant NSObject
s" NSString" objc_class Constant NSString
s" NSNumber" objc_class Constant NSNumber
s" NSAutoreleasePool" objc_class Constant NSAutoreleasePool

s" alloc" objc_sel Constant sel_alloc
s" init" objc_sel Constant sel_init
s" retain" objc_sel Constant sel_retain
s" release" objc_sel Constant sel_release
s" autorelease" objc_sel Constant sel_autorelease

: sel" \ usage example: sel" stringWithString:"
POSTPONE s"
POSTPONE ftoc
POSTPONE objc_registerName ; immediate

: class" \ usage example: class" NSString"
POSTPONE s"
POSTPONE ftoc
POSTPONE objc_getClass ; immediate

: NSStringFromForth ( addr u -- ptr )
ftoc NSString sel" stringWithCString:" rot objc_msgSend1 ;

: @" \ replacement for s" which produces an NSString instead
POSTPONE s"
POSTPONE NSStringFromForth ; immediate

: NSAutoreleasePoolPush ( -- ptr )
NSAutoreleasePool sel_alloc objc_msgSend0 sel_init objc_msgSend0 ;
: objc_release ( ptr -- )
sel_release objc_msgSend0 drop ;
: objc_retain ( ptr -- ptr )
sel_retain objc_msgSend0 ;
: objc_autorelease ( ptr -- ptr )
sel_autorelease objc_msgSend0 ;

: ndrop ( x1 ... xu u -- )
0 ?DO drop LOOP ;

: objc_msgSend ( x1 ... xu u obj sel -- result )
rot 2 + >r swap 2dup r> -rot \ add obj and sel to arguments
>r >r
dup cells 7 + 7 invert and 128 + \ this code was translated from the macro marg_malloc
dup >r allocate throw ( args... argc ptr )
dup >r 128 + dup rot cells + swap ( args... ptrend ptrstart ) +DO
i !
cell +LOOP \ advance one cell
r> dup r> r> r> ( argptr argptr argc sel obj )
swap 2swap swap objc_msgSendv swap free drop ;

: callocate \ allocate and erase space
dup allocate throw dup rot erase ;

: -@> ( "classname" superclass -- classptr methodliststruct )
parse-word ftoc swap ( move class name to back of stack for later use )

10 cells callocate \ allocate memory for class structure

10 cells callocate \ allocate memory for meta class structure
( classname superclass classstruct metaclassstruct )
2 pick
BEGIN \ search root class
dup cell + ( ptr to superclass ) @ dup 0= IF drop 1 ELSE nip 0 ENDIF
UNTIL ( classname superclass classstruct metaclassstruct rootclass )
@ over ! \ the meta class uses the root class' isa ptr as its isa

2 pick @ over cell + ! \ metaclass->super_class = super_class->isa

cell callocate over 7 cells + ! \ store (empty) class method list

$2 ( CLS_META ) over 4 cells + ( ptr to info ) !
( classname superclass classstruct metaclassstruct )
3 pick over 2 cells + ( ptr to name ) !

over ! \ store metaclass as isa-ptr of class structure
( classname superclass classstruct )
rot over 2 cells + ! \ store class name
( superclass classstruct )
dup -rot cell + ! \ store super_class
cell callocate over 7 cells + ! \ store (empty) method list
dup objc_addClass
2 cells callocate ( allocate struct objc_method_list ) ;

: @method ( "name" "signature" methodliststruct callback -- methodstruct )
parse-word objc_sel parse-word ftoc ( methodliststruct callback selector signature )
2swap swap ( selector signature callback methodliststruct )
dup cell + @ ( count ) dup 1+ dup 3 cells * 2 cells + >r rot r> resize throw \ add area for new method
( selector signature callback oldmethodcount newmethodcount newmethodliststruct )
swap over cell + ! \ increment method count
( selector signature callback oldmethodcount newmethodliststruct )
swap 3 cells * 2 cells + over + ( calculate offset in list )
( selector signature callback newmethodliststruct newmethodstruct )
rot over 2 cells + ! \ store method implementation
( selector signature newmethodliststruct newmethodstruct )
rot over cell + ! \ store method signature
( selector newmethodliststruct newmethodstruct )
>r swap r> ! \ store method selector
;

: @end ( classstruct methodliststruct -- )
objc_addMethods ;

callback IBAction (void) ptr ptr ptr callback; \ parameters: self _cmd sender

\ Example Usage:
\
\ NSObject -@> ForthClass
\ ( 4 cells instance myInstanceVariable )
\ 
\ :noname drop drop drop @" mmoooo!" NSLog0 ; IBAction test
\ test @method test: v@:@
\
\ @end
